/*
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	$Id: winquake.rc,v 1.6 2007/01/07 18:11:03 disconn3ct Exp $
*/
//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "winresrc.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_DIALOG2 DIALOGEX 50, 20, 317, 227
STYLE WS_MINIMIZEBOX | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Mvdsv - QW server"
CLASS "mvdsv"
FONT 10, "Fixedsys"
BEGIN
    EDITTEXT        IDC_EDIT2,6,187,306,13,ES_AUTOHSCROLL | WS_GROUP
    DEFPUSHBUTTON   "OK",IDC_OK,303,211,9,10,NOT WS_VISIBLE
    PUSHBUTTON      "QUIT",IDC_QUIT,262,207,50,14,WS_GROUP
    EDITTEXT        IDC_EDIT1,6,5,306,175,ES_MULTILINE | ES_AUTOVSCROLL | 
                    ES_READONLY | ES_WANTRETURN | NOT WS_BORDER | WS_VSCROLL | 
                    NOT WS_TABSTOP,WS_EX_CLIENTEDGE
    PUSHBUTTON      "CLEAR",IDC_CLEAR,6,207,50,14
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO DISCARDABLE 
BEGIN
    IDD_DIALOG2, DIALOG
    BEGIN
        LEFTMARGIN, 6
        RIGHTMARGIN, 312
        TOPMARGIN, 5
        BOTTOMMARGIN, 221
    END
END
#endif    // APSTUDIO_INVOKED


#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
//BEGIN
//    "#include ""afxres.h""\r\n"
//    "\0"
//END

//3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_ICON2               ICON    DISCARDABLE     "qwcl2.ico"
#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

