/*

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __RESOURCE_H__
#define __RESOURCE_H__

//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by winquake.rc
//
#define IDS_STRING1                     1
#define IDI_ICON2                       1
#define IDD_DIALOG1                     108
#define IDD_PROGRESS                    109
#define IDB_QWBITMAP                    112
#define IDB_ZQBITMAP                    115
#define IDB_BITMAP1                     117
#define IDD_DIALOG2                     118
#define IDR_MENU1                       121
#define IDC_PROGRESS                    1000
#define IDC_CLEAR                       1001
#define IDC_QUIT                        1002
#define IDC_EDIT1                       1003
#define IDC_EDIT2                       1004
#define IDC_OK                          1005
#define IDC_RESTORE                     1006
#define ID_MENU                         40003
#define ID_MENU_RESTORE                 40004
#define ID_MENU_ABOUT                   40005
#define ID_MENU_QUIT                    40006

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NO_MFC                     1
#define _APS_NEXT_RESOURCE_VALUE        122
#define _APS_NEXT_COMMAND_VALUE         40007
#define _APS_NEXT_CONTROL_VALUE         1002
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif

#endif /* !__RESOURCE_H__ */
